#============================================================================
#====================  Generic Procedures for expressions            ========
#====================  Author: B. Alpers, FH Aalen, 30.9.00          ========
#============================================================================
#============================================================================


#============================================================================
#======== Find name in correct set  =========================================
#============================================================================

find_name:=proc(name,menge)
	local i;
        if type(menge,set)=false then RETURN([1,[]]); fi; # no set
	for i from 1 to nops(menge) do
	    if op(1,menge[i])=name then RETURN([0,op(2,menge[i])]); fi; od;
	RETURN([2,[]]);   # not found
end:


#============================================================================
#======== Check expression for correctness  =================================
#============================================================================

G_check_expression:=proc(expression,correct_expression)

#=== check whether the expression is equal to the correct expression.
#=== If so, the routine returns 0, else 1.

if simplify(expression-correct_expression)=0 then RETURN(0)
else RETURN(1);
fi:

end:

#============================================================================
#====================  Find error in wrong expression		     ========
#============================================================================


G_find_error_expression:=proc(expression,correct_expression)

#=== Return - if possible - the kind of error, the wrong part of the expression,
#=== The error types currently implemented are:
#=== 0 = NO_ERROR
#=== 1 = NOT_IDENTIFIED (none of the below)
#=== 2 = WRONG_SYMBOLS (set of wrong symbols)
#=== 3 = SIGN_ERROR (set of terms in the sum with wrong signs) - for one term or terms in a sum
#=== 4 = WRONG_FACTORS (set of terms in the sum with wrong factors other than -1)
#=== 5 = WRONG_TERMS (set of wrong terms other than in 3 and 4) - for terms in a sum
#=== 6 = MISSING_TERMS (set of missing terms)
#=== 7 = MISSING_SYMBOLS (set of missing symbols)
#=== 8 = WRONG_TYPE: type of expression is wrong (e.g. `+` instead of `*`)


local 	NO_ERROR, NOT_IDENTIFIED, WRONG_SYMBOLS, SIGN_ERROR, WRONG_FACTORS,
	WRONG_TERMS, MISSING_TERMS, MISSING_SYMBOLS, WRONG_TYPE, 
	setlist, 
	compare_product_result, compare_sum_result, compare_result;

# Initialize error type variables
NO_ERROR:=0;
NOT_IDENTIFIED:=1;
WRONG_SYMBOLS:=2;
SIGN_ERROR:=3;
WRONG_FACTORS:=4;
WRONG_TERMS:=5;
MISSING_TERMS:=6;
MISSING_SYMBOLS:=7;
WRONG_TYPE:=8;

#=== check for errors

if G_check_expression(expression,correct_expression)=0 then
   RETURN([0,{}]);
fi;


#=== first check whether the equation contains a wrong or a missing symbol

setlist:=G_compare_symbol_sets(expression,correct_expression);
if setlist[1] <> {} then 
	RETURN([WRONG_SYMBOLS, setlist[1]]);
fi;

if setlist[2] <> {} then 
	RETURN([MISSING_SYMBOLS, setlist[2]]);
fi;

#=== next check whether the there is a simple sign error

if expression=-correct_expression then RETURN([SIGN_ERROR, {expression}]); fi;


#=== next check products or sums for wrong operands

if whattype(correct_expression)=`*`  then
   if whattype(expression)<>`*`
   	then RETURN([WRONG_TYPE,{whattype(expression)}]);
   else compare_product_result:=
		G_compare_product_expressions(factor(expression),correct_expression);
	if nops(compare_product_result[1])=1 and 
		type(op(1,compare_product_result[1]),numeric)=true
   		then RETURN([WRONG_FACTORS,compare_product_result[1]]); fi;
        if compare_product_result[1]<>{} 
		then RETURN([WRONG_TERMS,compare_product_result[1]]);
	elif compare_product_result[2]<>{} 
		then RETURN([MISSING_TERMS,compare_product_result[2]]);
	fi;
   fi;
fi;
      

if whattype(correct_expression)=`+`  then
   if whattype(expression)<>`+`
   	then RETURN([WRONG_TYPE,{whattype(expression)}]);
   else compare_sum_result:=
		G_compare_sum_expressions(expand(expression),correct_expression);
	if nops(compare_sum_result[1])=1 and nops(compare_sum_result[2])=1
   	   and type(op(1,compare_sum_result[1])/op(1,compare_sum_result[2]),numeric)=true
   	   then RETURN([WRONG_FACTORS,compare_sum_result[1]]); fi;
        if compare_sum_result[1]<>{} 
		then RETURN([WRONG_TERMS,compare_sum_result[1]]);
	elif compare_sum_result[2]<>{} 
		then RETURN([MISSING_TERMS,compare_sum_result[2]]);
	fi;
   fi;
fi;


#=== nothing specific identified, give up

RETURN([NOT_IDENTIFIED,{}]); 
end:



#============================================================================
#=====================  Extract symbol set from any expression       ========
#============================================================================


G_extract_symbol_set:=proc(any_expression)

# any_expression can be any expression, e.g. a system of equations

local i,symbol_set;
symbol_set:={};

if whattype(any_expression)=symbol or whattype(any_expression)=string #letzteres aus Kompatibilitaet mit R4
		then symbol_set:={any_expression};
elif type(any_expression,complex)=false then 
     for i from 1 to nops(any_expression) do
 	symbol_set:=symbol_set union G_extract_symbol_set(op(i,any_expression));
     od;
fi;
symbol_set;
end:




#============================================================================
#=====================  compare symbol sets	                     ========
#============================================================================

G_compare_symbol_sets:=proc(expr1,expr2)

#=== compare the symbol sets of two expressions where the first expression
#=== is assumed to be the correct one. The procedure returns
#=== a list with two sets of symbols: the first contains wrong symbols in system2,
#=== the second contains missing symbols

local set1,set2,result_list,i;

result_list:=array(1..2);
set1:=G_extract_symbol_set(expr1);
set2:=G_extract_symbol_set(expr2);

result_list[1]:=set1 minus set2; # should be empty set if system1 contains correct
				# symbols
result_list[2]:=set2 minus set1; # contains those symbols missing in set1

result_list;

end:



#============================================================================
#=====================  compare product expressions	             ========
#============================================================================

G_compare_product_expressions:=proc(expr1,expr2)

#=== compare two product expressions where the second expression
#=== is assumed to be the correct one. The procedure returns
#=== a list with two sets of operands: the first contains wrong operands in expr1,
#=== the second contains missing operands in expr1.
#=== if one of the expressions is not a product expression (i.e. not of type *)
#=== then the procedure returns 1

local set1,set2,result_list,i;

if not (whattype(expr1)=`*` and whattype(expr2)=`*`) then RETURN(1); fi;

result_list:=array(1..2);
set1:={seq(op(i,expr1),i=1..nops(expr1))};
set2:={seq(op(i,expr2),i=1..nops(expr2))};

result_list[1]:=set1 minus set2; # should be empty set if expr2 contains only 
				# correct factors
result_list[2]:=set2 minus set1; # contains those factors missing in expr2

result_list;

end:

#============================================================================
#=====================  compare sum expressions	                     ========
#============================================================================

G_compare_sum_expressions:=proc(expr1,expr2)

#=== compare two sum expressions where the second expression
#=== is assumed to be the correct one. The procedure returns
#=== a list with two sets of symbols: the first contains wrong operands in expr1,
#=== the second contains missing operands in expr1.
#=== if one of the expressions is not a sum expression (i.e. not of type +)
#=== then the procedure returns 1

local set1,set2,result_list,i;

if not (whattype(expr1)=`+` and whattype(expr2)=`+`) then RETURN(1); fi;

result_list:=array(1..2);
set1:={seq(op(i,expr1),i=1..nops(expr1))};
set2:={seq(op(i,expr2),i=1..nops(expr2))};

result_list[1]:=set1 minus set2; # should be empty set if expr1 contains correct
				# operands
result_list[2]:=set2 minus set1; # contains those operands missing in expr1

result_list;

end: